/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2005 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <cmath>

#include <boost/variant/get.hpp>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_hyperg.h>
#include <gsl/gsl_sf_result.h>
#include <gsl/gsl_sf_gamma.h>

#include "libtkcommon/debug.h"
#include "libtkquery/parse.h"
#include "analyser.h"

using namespace Tk::Analysis;

const Tk::Rational*
PositionFunction:: operator()(const Position &pos) const
{
  PositionFunction::const_iterator pos_iter( find(pos) );
  if (pos_iter == end())
    return 0;
  
  return &pos_iter->second;
}

Tk::Rational
PositionFunction:: calc_rss(const Probability &prob) const
{
  /*  Rational sample_sum(0);
  for (PositionFunction::const_iterator pos_iter( begin() );
       pos_iter != end();
       ++pos_iter)
    sample_sum += pos_iter->second;
  
  Rational sample_mean( sample_sum / static_cast<Rational>(size()) );
  */
  //  Rational sample_mean( calc_mean() );
  
  
  Tk::Rational difference;
  Tk::Rational square;
  Tk::Rational SS_E(0);
  for (PositionFunction::const_iterator pos_iter( begin() );
       pos_iter != end();
       ++pos_iter)
    {
      difference = pos_iter->second - prob;
      square = difference * difference;
      SS_E += square;
    }
  
  return SS_E;
}

Tk::Rational
PositionFunction:: calc_mean() const
{
  Tk::Rational sum(0);
  for (PositionFunction::const_iterator iter( begin() );
       iter != end();
       ++iter)
    sum += iter->second;
  
  return sum / static_cast<Tk::Rational>( size() );
}

Tk::Rational
PositionFunction:: calc_variance(const Tk::Rational &mean) const
{
  Tk::Rational sum(0);
  Tk::Rational difference;
  Tk::Rational square;
  
  for (PositionFunction::const_iterator iter( begin() );
       iter != end();
       ++iter)
    {
      difference = iter->second - mean;
      square = difference * difference;
      sum += square;
    }
  return sum / static_cast<Tk::Rational>( size() );
}

Tk::Real
PositionFunction:: calc_skewness(const Tk::Rational &mean, const Tk::Rational &variance) const
{
  Tk::Rational sum(0);
  Tk::Rational difference;
  Tk::Rational cube;
  
  if (size() == 1)
    return 0.0;
  
  for (PositionFunction::const_iterator iter( begin() );
       iter != end();
       ++iter)
    {
      difference = iter->second - mean;
      cube = difference * difference * difference;
      sum += cube;
    }
  
  return ( sqrt( static_cast<Tk::Real>(size()) ) * static_cast<Tk::Real>(sum) ) / static_cast<Tk::Real>(pow(variance.get_d(), 2.0/3.0));
}

OffsetStatisticsSet:: OffsetStatisticsSet(const OffsetProbabilities &probs,
					  const PositionOffsetDistributions &offset_distributions,
					  const PositionCounts &position_counts)
  : _position_totals(position_counts)
{
  /* Offset list and theoretical probabilities */
  for (OffsetProbabilities::const_iterator prob_iter( probs.begin() );
       prob_iter != probs.end();
       ++prob_iter)
    {
      _offsets.push_back(prob_iter->first);
      _offset_stats[prob_iter->first]._probability = prob_iter->second;
    }
  
  /* Distributions and counts*/
  OffsetDistribution offset_position_counts;
  for (PositionCounts::const_iterator pos_iter( position_counts.begin() );
       pos_iter != position_counts.end();
       ++pos_iter)
    {
      const OffsetDistribution &offset_distribution( offset_distributions.find(pos_iter->first)->second );
      for (OffsetDistribution::const_iterator offset_iter( offset_distribution.begin() );
	   offset_iter != offset_distribution.end();
	   ++offset_iter)
	{
	  /* Distributions */
	  Tk::Rational div(static_cast<Tk::Rational>(offset_iter->second) / pos_iter->second);
	  PositionFunction::value_type pos_func_row(pos_iter->first, div);
	  _offset_stats[offset_iter->first]._position_function.insert(pos_func_row);
	  
	  /* Counts */
	  _offset_stats[offset_iter->first]._position_counts[pos_iter->first] = offset_iter->second;
	}
    }
  
  /* RSS, mean and variance */
  for (OffsetStatisticsMap::iterator offset_iter( _offset_stats.begin() );
       offset_iter != _offset_stats.end();
       ++offset_iter)
    {
      const PositionFunction &pos_func( offset_iter->second._position_function );
      if (pos_func.size() < 1)
	continue;
      
      OffsetStatistics &stats( offset_iter->second );
      stats._mean = pos_func.calc_mean();
      stats._rss = pos_func.calc_rss(stats._probability);
      stats._variance = pos_func.calc_variance(stats._mean);
      stats._skewness = pos_func.calc_skewness(stats._mean, stats._variance);
    }
}

Statistics:: Statistics(Db::Database &database,
			const IntentCode &intent_code,
			const OffsetProbabilities &offset_probs,
			const Db::IdList &experiments)
  : _intent_code(intent_code),
    _n(0),
    _position_min(0),
    _position_max(0),
    _experiment_count(0)
{ 
  Offset results_row;
  Offset offset;
  int screen_pos;
  Tk::Real screen_width;
  Position pos;
  PositionCounts pos_counts;
  PositionOffsetDistributions pos_offsets;
  
  /*
   * Compile stats
   */
  for (Db::IdList::const_iterator ex_iter(experiments.begin());
       ex_iter != experiments.end();
       ++ex_iter)
    {
      ExperimentStatistics ex;
      Count ex_n(0);
      Position ex_pos_sum(0);
      Db::Id ex_id(*ex_iter);
      
      ++_experiment_count;
      
      screen_pos = 0;
      pos = 0;
      
      screen_width = database.get_property("experiment", ex_id, "screen width").get_datum();
     
      Db::DataSetPtr results( database.get_results(ex_id) );
      
      bool results_valid( (*results)->next() );
      if (!results_valid)
	{
	  String msg(L"No results for experiment ");
	  msg += String(ex_id);
	  throw std::runtime_error(static_cast<std::string>(msg));
	}
      
      for(;
	  !(*results)->isAfterLast();
	  (*results)->next())
	{
	  offset = (*results)->getInt(1);
	  
	  screen_pos += offset;	
	  pos = static_cast<Position>(screen_pos) / static_cast<Position>(screen_width);
	  
	  /* overall */
	  ++_n;
	  if (_position_max < pos)
	    _position_max = pos;
	  else if (_position_min > pos)
	    _position_min = pos;
	  
	  /* offset */
	  ++(pos_counts[pos]);
	  ++(pos_offsets[pos][offset]);
	  
	  /* experiment */
	  ++ex_n;
	  ex_pos_sum += pos;
	}
      
      /* complete experiment stats */
      ex._position_mean = ex_pos_sum / ex_n;
      _experiments.push_back(ex);
    }
  
  /* complete overall stats */
  _offsets = OffsetStatisticsSet(offset_probs,
				 pos_offsets,
				 pos_counts);
  
  /*
   * Calculate mean and variance of experiments' position mean
   */
  Position square;
  Position difference;
  
  /* grand mean */
  Position ex_pos_sum(0);
  for(ExperimentStatisticsSet::iterator iter(_experiments.begin());
      iter != _experiments.end();
      ++iter)
    {
      ex_pos_sum += iter->_position_mean;
    }
  
  _experiment_position_mean = ex_pos_sum / static_cast<Tk::Rational>(_experiment_count);
  
  /* variance */
  ex_pos_sum = 0;
  for(ExperimentStatisticsSet::iterator iter(_experiments.begin());
      iter != _experiments.end();
      ++iter)
    {
      difference = iter->_position_mean - _experiment_position_mean;
      square = difference * difference;
      ex_pos_sum += square;
    }
  
  _experiment_position_variance = ex_pos_sum / static_cast<Tk::Rational>(_experiment_count);
}

AlgorithmStatistics:: AlgorithmStatistics(Db::Database &database,
					  const IntentCodeSet &intent_codes,
					  const String &algo_spec,
					  const IntentExperiments &experiments)
  : _algorithm_spec(algo_spec)
{
  Algorithm algo(algo_spec);
  _offset_probabilities = algo.get_probabilities();
  
  for (IntentCodeSet::const_iterator iter(intent_codes.begin());
       iter != intent_codes.end();
       ++iter)
    {
      IntentExperiments::const_iterator intent_exs(experiments.find(*iter));
      if (intent_exs == experiments.end())
	{
	  std::wcerr << L"Experiments for intent code '"
		     << String(*iter)
		     << "' not found"
		     << std::endl;
	  continue;
	}
      
      _intent_stats.push_back(Statistics(database, intent_exs->first, _offset_probabilities, intent_exs->second));
    }
}


TraceStatistics:: TraceStatistics (Db::Database &database,
				   unsigned trace_count,
				   const AlgorithmExperiments &experiments,
				   const IntentCodeSet &intent_codes)
  : _trace_count(trace_count),
    _intent_codes(intent_codes)
{
  for (AlgorithmExperiments::const_iterator iter( experiments.begin() );
       iter != experiments.end();
       ++iter)
    {
      const String &algo_spec(iter->first);
      const IntentExperiments &algo_experiments(iter->second);
      
      _algorithm_stats.push_back(AlgorithmStatistics(database, _intent_codes, algo_spec, algo_experiments));
    }
}

static void
gsl_error_handler (const char * reason,
		   const char * file,
		   int line,
		   int gsl_errno)
{
  std::ostringstream ostr;
  ostr << file
       << ": "
       << line
       << ": "
       << reason
       << " (GSL error code "
       << gsl_errno
       << ")";
  throw std::runtime_error(ostr.str());
}


TTest:: TTest(const Statistics &stats_a, const Statistics &stats_b)
  : _t(0.0),
    _df( stats_a._experiment_count + stats_b._experiment_count - 2 ),
    _p(0.0)
{
  // from http://www.socialresearchmethods.net/kb/stat_t.php
  /*  Tk::Rational term_a(stats_a._experiment_position_variance / stats_a._experiment_position_mean);
  Tk::Rational term_b(stats_b._experiment_position_variance / stats_b._experiment_position_mean);
  Tk::Real std_error( sqrt(static_cast<Tk::Real>(term_a + term_b)) );
  
  Tk::Rational mean_difference(stats_a._experiment_position_mean - stats_b._experiment_position_mean);
  Tk::Rational abs_mean_difference(abs(mean_difference));
  
  _t = abs_mean_difference / std_error;*/
  
  Tk::Rational mean_difference( stats_a._experiment_position_mean - stats_b._experiment_position_mean );
  
  Tk::Rational sqrt_dividend( (stats_a._experiment_count-1)*stats_a._experiment_position_variance
			  + (stats_b._experiment_count-1)*stats_b._experiment_position_variance );
  Tk::Rational sqrt_divisor(_df);
  Tk::Rational sqrt_second_term( static_cast<Tk::Rational>(1)/stats_a._experiment_count
			     + static_cast<Tk::Rational>(1)/stats_b._experiment_count );
  Tk::Real std_error( sqrt(static_cast<Tk::Real>( (sqrt_dividend/sqrt_divisor) * sqrt_second_term )) );
  
  _t = mean_difference / std_error;
  
  // from http://www.nrbook.com/a/bookcpdf/c6-4.pdf (my god what a trek)
  gsl_set_error_handler(gsl_error_handler);
  unsigned long v(_df.get_ui());
  _p = gsl_sf_beta_inc(0.5*v, 0.5, v / (v + static_cast<mpf_class>(_t*_t).get_d()));
}

Analyser:: Analyser(Db::Database &database,
		    const String &query_file)
{
  Db::IdList experiments( Query::get_experiments(&database, query_file) );
  
  std::wcout << String(experiments.size()) << L" experiments..." << std::flush;

  // Sort experiments by trace count, algorithm and intent code
  typedef std::map<unsigned, IntentCodeSet> TraceIntentCodes;
  typedef std::map<unsigned, AlgorithmExperiments> TraceExperiments;
  
  TraceIntentCodes intent_codes;
  TraceExperiments sorted_experiments;
  
  for (Db::IdList::iterator e( experiments.begin() );
       e != experiments.end();
       ++e)
    {
      Db::Id ex_id( *e );
      unsigned trace_count( database.get_property("experiment", ex_id, "trace count").get_datum() );
      String algo_spec( database.get_property("experiment", ex_id, "algorithm spec").get_datum() );
      IntentCode intent_code( database.get_property("experiment", ex_id, "intent code").get_datum() );
      
      intent_codes[trace_count].insert(intent_code);
      
      sorted_experiments[trace_count][algo_spec][intent_code].push_back(ex_id);
    }
  
  for (TraceExperiments::iterator titer( sorted_experiments.begin() );
       titer != sorted_experiments.end();
       ++titer)
    {
      unsigned trace_count( titer->first );
      AlgorithmExperiments &algo_experiments( titer->second );
      
      _statistics.push_back(TraceStatistics(database, trace_count, algo_experiments, intent_codes[trace_count]));
    }
}

