/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006, 2007, 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <time.h>

#include <sstream>

#include <boost/variant/get.hpp>
#include <boost/variant/static_visitor.hpp>
#include <boost/variant/apply_visitor.hpp>

#include <odbc++/setup.h>
#include <odbc++/drivermanager.h>
#include <odbc++/databasemetadata.h>

#include "libtkcommon/tkintl.h"
#include "libtkcommon/mbs.h"
#include "libtkcommon/exceptions.h"
#include "libtkcommon/tkvalue.h"
#include "libtkcommon/debug.h"
#include "libtkdb/database.h"
#include "libtkdb/value.h"
#include "libtkquery/query.h"

using namespace Tk::Db;

UnknownType:: UnknownType(const Tk::String &type_name)
  : std::logic_error(std::string(_("Unknown type '")) +
		     static_cast<std::string>(type_name) +
		     _("'")),
    _type_name(type_name)
{}

UnknownType:: ~UnknownType()
  throw ()
{
}

const Tk::String &
UnknownType:: get_type_name() const
{
  return _type_name;
}

PropertyNotSet:: PropertyNotSet(const Tk::String &domain, const Id &entity_id, const Tk::String &property_name)
  : std::runtime_error(std::string(_("Property `"))
		       + static_cast<std::string>(property_name)
		       + _("' not set for entity ")
		       + static_cast<std::string>(Tk::String(entity_id))
		       + _(" in domain `")
		       + static_cast<std::string>(domain)
		       + "'"
		       )
{}



Database:: Database (const Tk::String &data_source_name,
		     const Tk::String &username,
		     const Tk::String &password)
  : _connection(odbc::DriverManager::getConnection(data_source_name,
						   username,
						   password)
		)
{
  //WDEBUG_OUT << L"Getting database connection metadata..." << std::endl;
  odbc::DatabaseMetaData *metadata( _connection->getMetaData() );
  //WDEBUG_OUT << L"...got database connection metadata" << std::endl;
  
  std::wcout << _(L"Database product: ")
	     << Tk::String(metadata->getDatabaseProductName())
	     << _(L"; version: ")
	     << Tk::String(metadata->getDatabaseProductVersion())
	     << std::endl
	     << _(L"Driver: ")
	     << Tk::String(metadata->getDriverName())
	     << _(L"; version: ")
	     << Tk::String(metadata->getDriverVersion())
	     << std::endl;
  
  //WDEBUG_OUT << L"Initialised Database" << std::endl;
}

Database:: ~Database()
{
}

Statement *
Database:: new_statement()
{
  return _connection->createStatement(odbc::ResultSet::TYPE_FORWARD_ONLY,
				      odbc::ResultSet::CONCUR_READ_ONLY);
}

void
Database:: execute_statement(const Tk::String &sql)
{
  StatementPtr stmt( new_statement() );
  stmt->execute(sql);
}

DataSet *
Database:: execute_query(const Tk::String &sql)
{
  return new DataSet( new_statement()->executeQuery(sql) );
}

PStatement *
Database:: prepare_statement(const Tk::String &sql)
{
  return _connection->prepareStatement(sql,
				       odbc::ResultSet::TYPE_FORWARD_ONLY,
				       odbc::ResultSet::CONCUR_READ_ONLY);
}

void
Database:: get_table_info(const Tk::String &name)
{
  odbc::DatabaseMetaData *metadata( _connection->getMetaData() );
  std::auto_ptr<odbc::ResultSet> result( metadata->getColumns(String(), String(), name, String()) );
  
  FieldInfo fields;
  for (result->next(); !result->isAfterLast(); result->next())
    {
      odbc::Types::SQLType odbc_type( static_cast<odbc::Types::SQLType>(result->getShort(5)) );
      if (!odbc_type)
	{
	  WDEBUG_OUT << L"Zero type for column '"
		     << Tk::String(result->getString(4))
		     << L"'; driver-specific type name: '"
		     << Tk::String(result->getString(6))
		     << L"'"
		     << std::endl;
	}
      fields.insert(FieldInfo::value_type(result->getString(4), odbc_type));
    }
  
  _table_info.insert(TableInfo::value_type(name, fields));
}

void
Database:: check_table_info(const Tk::String &name)
{
  if (_table_info.find(name) == _table_info.end())
    get_table_info(name);
}

Type
Database:: get_column_type(const Tk::String &table, const Tk::String &column)
{
  check_table_info(table);
  
  const FieldInfo &fields(_table_info[table]);
  
  FieldInfo::const_iterator iter = fields.find(column);
  if (iter == fields.end())
    throw std::runtime_error(std::string("Could not find column ") +
			     static_cast<std::string>(table) + "." +
			     static_cast<std::string>(column));
  return iter->second;
}

DataSet *
Database:: get_table_data(const Tk::String &table, const Tk::StringList &columns)
{
  std::wostringstream query;
  query << L"select ";
  Tk::StringList::const_iterator iter, next;
  for (iter = columns.begin();
       iter != columns.end();
       iter = next)
    {
      next = iter;
      ++next;
      
      query << *iter;
      if (next != columns.end())
	query << L",";
    }
  query << L" from "
	<< table
	<< L" t";
  
  Tk::String sql(query.str());
  Statement *stmt(new_statement());
  DataSet *result( new DataSet(stmt->executeQuery(sql)) );
  
  return result;
}

PStatement *
Database:: get_table_insert_statement(const Tk::String &table, const Tk::StringList &columns)
{
  std::wostringstream query;
  std::wostringstream values;

  query << L"insert into " << table << L"(";
  values << L") values (";

  Tk::StringList::const_iterator iter, next;
  for(iter = columns.begin();
      iter != columns.end();
      iter = next)
    {
      next = iter; ++next;
      
      query << *iter;
      values << L"?";
      
      if(next != columns.end())
	{
	  query << L", ";
	  values << L", ";
	}
    }
  query << values.str();
  query << L")";
  
  return prepare_statement(query.str());
}


Id
Database:: get_last_insert_id()
{
  Tk::String sql;
  
  Tk::String db_product( _connection->getMetaData()->getDatabaseProductName() );
  WDEBUG_OUT << L"Database product name: " << db_product << std::endl;
  if (db_product == "SQLite")
    sql = "select last_insert_rowid()";
  else
    sql = "select last_insert_id()";

  Statement *stmt( new_statement() );
  DataSetPtr result( new DataSet(stmt->executeQuery(sql)) );
  
  bool valid( (*result)->next() );
  if (!valid)
    throw std::runtime_error("Could not get last insert id");
  
  return (*result)->getLong(1);
}

void
Database:: property_type_get_details (const Tk::String &domain,
				      const Tk::String &property_name,
				      Id *property_type_id,
				      Tk::String *property_value_type_name,
				      Tk::String *property_value_table_name)
{
  const static Tk::String
    sql(" select    pt.property_type_id"
	" ,         pvt.name"
	" ,         pvt.value_table"
	" from      property_type pt"
	" ,         property_value_type pvt"
	" where     pt.domain = ?"
	" and       pt.name = ?"
	" and       pvt.property_value_type_id = pt.property_value_type_id"
	);

  PStatement *query( prepare_statement(sql) );
  
  query->setString(1, domain);
  query->setString(2, property_name);
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw UnknownType(property_name);
  
  if (property_type_id)
    *property_type_id = result->getLong(1);
  if (property_value_type_name)
    *property_value_type_name = result->getString(2);
  if (property_value_table_name)
    *property_value_table_name = result->getString(3);
}

Id
Database:: insert_property(const Id &property_type_id)
{
  Tk::StringList columns;
  columns.push_back("property_type_id");
  
  PStatementPtr stmt( get_table_insert_statement("property", columns) );
  
  stmt->setLong(1, property_type_id);
  stmt->execute();
  
  return get_last_insert_id();
}

void
Database:: insert_property_value_data (const Id &property_id,
				       const Tk::String &table_name,
				       const Data::Value &value)
{
  Tk::StringList columns;
  columns.push_back("property_id");
  columns.push_back("property_value");

  PStatementPtr stmt( get_table_insert_statement(table_name, columns) );
  
  stmt->setLong(1, property_id);
  
  value.set_parameter(stmt.get(), 2);
  
  stmt->execute();
}

void
Database:: update_property_value_data (const Id &property_id,
				       const Tk::String &table_name,
				       const Data::Value &value)
{
  std::wostringstream query;
  query << L" update " << table_name
	<< L" set    property_value = ?"
	<< L" where  property_id    = ?"
    ;

  PStatementPtr stmt( prepare_statement(query.str()) );
  
  value.set_parameter(stmt.get(), 1);
  stmt->setLong(2, property_id);
  
  stmt->execute();
}

void
Database:: insert_domain_property(const Tk::String &domain,
				  const Id &entity_id,
				  const Id &property_id)
{
  Tk::StringList columns;
  columns.push_back(domain + Tk::String("_id"));
  columns.push_back("property_id");
  
  PStatementPtr stmt( get_table_insert_statement(domain + Tk::String("_property"), columns) );
  
  stmt->setLong(1, entity_id);
  stmt->setLong(2, property_id);
  stmt->execute();
}


Id
Database:: set_property(const Tk::String &domain,
			const Id &entity_id,
			const Tk::String &property_name,
			const Data::Value &value)
{
  WDEBUG_OUT << L"Setting domain property '" << domain << L"." << property_name << "'.." << std::endl;
  Id property_type_id;
  Tk::String property_value_type_name;
  Tk::String property_value_table_name;
  
  WDEBUG_OUT << L"Getting type details..." << std::endl;
  property_type_get_details (domain,
			     property_name,
			     &property_type_id,
			     &property_value_type_name,
			     &property_value_table_name);
  WDEBUG_OUT << L"...got type details (type name: '"
	     << property_value_type_name
	     << L"'; table name: '"
	     << property_value_table_name
	     << L"')"
	     << std::endl;

  Id property_id;
  try
    {
      WDEBUG_OUT << L"Finding property..." << std::endl;
      property_id = find_property(domain, entity_id, property_name);
      WDEBUG_OUT << L"...found property (id: " << property_id << L")" << std::endl;
      
      WDEBUG_OUT << L"Updating property value data..." << std::endl;
      update_property_value_data(property_id,
				 property_value_table_name,
				 value);
      WDEBUG_OUT << L"...updated property value data" << std::endl;
    }
  catch (PropertyNotSet pns)
    {
      WDEBUG_OUT << L"...property not set; inserting new property..." << std::endl;
      property_id = insert_property(property_type_id);
      WDEBUG_OUT << L"...inserted property (id: " << property_id << L")" << std::endl;
      
      WDEBUG_OUT << L"Inserting domain property..." << std::endl;
      insert_domain_property(domain, entity_id, property_id);
      WDEBUG_OUT << L"...inserted domain property" << std::endl;
      
      WDEBUG_OUT << L"Inserting property value data..." << std::endl;
      insert_property_value_data(property_id,
				 property_value_table_name,
				 value);
      WDEBUG_OUT << L"...inserted property value data" << std::endl;
    }
  
  WDEBUG_OUT << L"..set domain property '" << domain << L"." << property_name << L"'" << std::endl;
  return property_id;
}


Id
Database:: lookup_location(const Tk::String &name)
{
  const static Tk::String sql("select location_id from location where name = ?");
  PStatement *stmt( prepare_statement(sql) );
  stmt->setString(1, name);
  
  DataSet result( stmt->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw std::runtime_error(std::string(_("Unknown location '")) + static_cast<std::string>(name) + _("'"));
  
  return result->getLong(1);
}

Id
Database:: insert_experiment(const Id &location_id,
			     const Random::Bits &randomness)
{
  Tk::StringList columns;
  columns.push_back("location_id");
  columns.push_back("randomness_bits");
  columns.push_back("randomness");
  
  PStatementPtr stmt( get_table_insert_statement("experiment", columns) );
  
  stmt->setLong(1, location_id);
  stmt->setLong(2, randomness.size());
  
  Random::BitIStream *randomness_stream( randomness.get_data() );
  stmt->setBinaryStream(3, randomness_stream, randomness.get_data_len());
  
  
  stmt->execute();
  
  delete randomness_stream;
  
  return get_last_insert_id();
}

void
Database:: experiment_insert_operators(const Id &experiment_id,
				       const IdList &operators)
{
  Tk::StringList columns;
  columns.push_back("experiment_id");
  columns.push_back("operator_id");
  
  PStatementPtr stmt( get_table_insert_statement("experiment_operator", columns) );
  
  stmt->setLong(1, experiment_id);
  
  for (IdList::const_iterator iter(operators.begin());
       iter != operators.end();
       ++iter)
    {
      stmt->setLong(2, *iter);
      stmt->execute();
    }
}

void
Database:: experiment_insert_result(const Id &experiment_id,
				    const Result &result)
{
  Tk::StringList columns;
  columns.push_back("experiment_id");
  columns.push_back("line_index");
  columns.push_back("offset_index");
  columns.push_back("offset");
  
  PStatementPtr stmt( get_table_insert_statement("experiment_result", columns) );
  
  stmt->setLong(1, experiment_id);
  
  unsigned ntraces( result.begin()->size() );
  unsigned line_index(0);
  unsigned offset_index;
  for (Result::const_iterator iter(result.begin());
       iter != result.end();
       ++iter, ++line_index)
    {
      stmt->setInt(2, line_index);
      
      for (offset_index = 0; offset_index < ntraces; ++offset_index)
	{
	  stmt->setInt(3, offset_index);
	  stmt->setInt(4, (*iter)[offset_index]);
	  
	  stmt->execute();
	}
    }
}

Id
Database:: insert_experiment(const Id &location_id,
			     const IdList &operators,
			     const Data::ValueMap &properties,
			     const Random::Bits &randomness,
			     const Result &result)
{
  DEBUG_OUT << "Inserting experiment.." << std::endl;
  
  DEBUG_OUT << "Inserting experiment table..." << std::endl;
  Id id( insert_experiment(location_id, randomness) );
  DEBUG_OUT << "...inserted experiment table" << std::endl;
  
  DEBUG_OUT << "Setting properties..." << std::endl;
  for (Data::ValueMap::const_iterator iter( properties.begin() );
       iter != properties.end();
       ++iter)
    set_property("experiment", id, iter->first, iter->second);
  DEBUG_OUT << "...set properties" << std::endl;
  
  
  DEBUG_OUT << "Inserting operators..." << std::endl;
  experiment_insert_operators(id, operators);
  DEBUG_OUT << "...inserted operators" << std::endl;

  DEBUG_OUT << "Inserting result..." << std::endl;
  experiment_insert_result(id, result);
  DEBUG_OUT << "...inserted result" << std::endl;
  
  DEBUG_OUT << "..inserted experiment (id: " << id << ")" << std::endl;
  return id;
}


Id
Database:: choose_location()
{
  Tk::String
    sql(" select   e.location_id"
	" from     experiment e"
	" ,        experiment_property ep"
	" ,        property p"
	" ,        property_type pt"
	" ,        property_value_datetime pv"
	" where    ep.experiment_id    = e.experiment_id"
	" and      p.property_id       = ep.property_id"
	" and      pv.property_id      = p.property_id"
	" and      pt.property_type_id = p.property_type_id"
	" and      pt.domain           = ?"
	" and      pt.name             = ?"
	" order by pv.property_value desc"
	);
  
  PStatement *query( prepare_statement(sql) );
  query->setString(1, Tk::String("experiment"));
  query->setString(2, Tk::String("time stamp"));
  
  DataSetPtr result( new DataSet(query->executeQuery()) );
  
  bool valid( (*result)->next() );
  if (!valid)
    {
      sql = "select location_id from location";
      result.reset( execute_query(sql) );
      
      valid = (*result)->next();
    }
  
  if (!valid)
    throw std::runtime_error(_("Could not choose location"));

  return (*result)->getLong(1);
}

Id
Database:: choose_operator()
{
  Tk::String
    sql(" select   eo.operator_id"
	" from     experiment e"
	" ,        experiment_operator eo"
	" ,        experiment_property ep"
	" ,        property p"
	" ,        property_type pt"
	" ,        property_value_datetime pv"
	" where    eo.experiment_id    = e.experiment_id"
        " and      ep.experiment_id    = e.experiment_id"
	" and      p.property_id       = ep.property_id"
	" and      pv.property_id      = p.property_id"
	" and      pt.property_type_id = p.property_type_id"
	" and      pt.domain           = ?"
	" and      pt.name             = ?"
	" order by pv.property_value desc"
	);
  
  PStatement *query( prepare_statement(sql) );
  query->setString(1, Tk::String("experiment"));
  query->setString(2, Tk::String("time stamp"));
  
  DataSetPtr result( new DataSet(query->executeQuery()) );
  
  bool valid( (*result)->next() );
  if (!valid)
    {
      sql = "select operator_id from operator";
      result.reset( execute_query(sql) );
      
      valid = (*result)->next();
    }
  
  if (!valid)
    throw std::runtime_error(_("Could not choose operator"));

  return (*result)->getLong(1);
}

Code
Database:: choose_intent(unsigned trace_count, const Tk::String &algorithm_spec, bool test)
{
  const static Tk::String
    sql(" select   i.intent_code"
	" from     intent_domain id"
	" ,        intent i"
	" where    id.trace_count = ?"
	" and      i.intent_domain_id = id.intent_domain_id"
	" order by ("
	"            select   count(*)"
	"            from     property_value_string pvs"
	"            ,        property p"
	"            ,        property_type pt"
	"            ,        experiment_property ep"
	"            ,        experiment e"
	"            ,        experiment_property epa"
	"            ,        property pa"
	"            ,        property_type pta"
	"            ,        property_value_string pvsa"
	"            ,        experiment_property epts"
	"            ,        property pts"
	"            ,        property_type ptts"
	"            ,        property_value_string pvsts"
	"            where    pvs.property_value        = i.intent_code"
	"            and      p.property_id             = pvs.property_id"
	"            and      pt.property_type_id       = p.property_type_id"
	"            and      pt.domain                 = ?"
	"            and      pt.name                   = ?"
	"            and      ep.property_id            = p.property_id"
	"            and      e.experiment_id           = ep.experiment_id"
	"            and      epa.experiment_id         = e.experiment_id"
	"            and      pa.property_id            = epa.property_id"
	"            and      pta.domain                = ?"
	"            and      pta.name                  = ?"
	"            and      pvsa.property_id          = pa.property_id"
	"            and      pvsa.property_value       = ?"
	"            and      epts.experiment_id        = e.experiment_id"
	"            and      pts.property_id           = epts.property_id"
	"            and      ptts.domain               = ?"
	"            and      ptts.name                 = ?"
	"            and      pvsts.property_id         = pts.property_id"
	"            and      pvsts.property_value      = ?"
	"	   )"
	);
  
  PStatement* query( prepare_statement(sql) );
  
  query->setInt(1, static_cast<int>(trace_count));
  query->setString(2, Tk::String("experiment"));
  query->setString(3, Tk::String("intent code"));
  query->setString(4, Tk::String("experiment"));
  query->setString(5, Tk::String("algorithm spec"));
  query->setString(6, algorithm_spec);
  query->setString(7, Tk::String("experiment"));
  query->setString(8, Tk::String("test"));
  query->setBoolean(9, test);
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    throw std::runtime_error(_("could not choose intent"));
  
  return result->getString(1);
}

IntentInfoList
Database:: get_intent_options(unsigned trace_count)
{
  const static Tk::String
    sql(" select   i.intent_code"
	" ,        i.description"
	" from     intent_domain id"
	" ,        intent i"
	" where    id.trace_count = ?"
	" and      i.intent_domain_id = id.intent_domain_id"
	);
  PStatement *query( prepare_statement(sql) );
  
  query->setInt(1, trace_count);
  
  DataSet result( query->executeQuery() );
  
  IntentInfoList info;
  for (result->next(); !result->isAfterLast(); result->next())
    info.push_back( IntentInfo(result->getString(1), result->getString(2)) );
  
  return info;
}

DataSet *
Database:: get_results(const Id &experiment_id)
{
  const static Tk::String
    sql(" select   offset"
	" from     experiment_result"
	" where    experiment_id = ?"
	" order by line_index"
	" ,        offset_index"
	);
  
  PStatement *stmt( prepare_statement(sql) );
  stmt->setLong(1, experiment_id);
  
  return new DataSet(stmt->executeQuery());
}

Id
Database:: find_property(const Tk::String &domain, const Id &entity_id, const Tk::String &property_name)
{
  Id property_type_id;
  
  property_type_get_details (domain,
			     property_name,
			     &property_type_id);
  
  std::wostringstream sqls;
  sqls << L" select   p.property_id"
       << L" from     " << domain << L" d"
       << L" ,        " << domain << L"_property dp"
       << L" ,        property p"
       << L" where    d." << domain << L"_id    = ?"
       << L" and      dp." << domain << L"_id   = d." << domain << L"_id"
       << L" and      p.property_id             = dp.property_id"
       << L" and      p.property_type_id        = ?"
    ;
  Tk::String sql(sqls.str());
  
  PStatement *query( prepare_statement(sql) );
  
  query->setLong(1, entity_id);
  query->setLong(2, property_type_id);
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    {
      throw PropertyNotSet(domain, entity_id, property_name);
    }
  
  return result->getLong(1);
}

Data::Value
Database:: get_property(const Tk::String &domain, const Id &entity_id, const Tk::String &property_name)
{
  Id property_type_id;
  Tk::String property_value_type_name;
  Tk::String property_value_table_name;
  
  property_type_get_details (domain,
			     property_name,
			     &property_type_id,
			     &property_value_type_name,
			     &property_value_table_name);
  
  std::wostringstream sqls;
  sqls << L" select   pv.property_value "
       << L" from     " << domain << L" d"
       << L" ,        " << domain << L"_property dp"
       << L" ,        property p"
       << L" ,        " << property_value_table_name << L" pv"
       << L" where    d." << domain << L"_id    = ?"
       << L" and      dp." << domain << L"_id   = d." << domain << L"_id"
       << L" and      p.property_id             = dp.property_id"
       << L" and      p.property_type_id        = ?"
       << L" and      pv.property_id            = p.property_id"
    ;
  Tk::String sql(sqls.str());
  
  PStatement *query( prepare_statement(sql) );
  
  query->setLong(1, entity_id);
  query->setLong(2, property_type_id);
  
  DataSet result( query->executeQuery() );
  
  bool valid( result->next() );
  if (!valid)
    {
      /*      Data::Datum d;
	      return Data::Value(Data::Boolean, d);*/
      throw PropertyNotSet(domain, entity_id, property_name);
    }
  
  return Data::Value(result, 1);
}

IdList
Database:: get_ids(PStatement &stmt)
{
  WDEBUG_OUT << L"getting ids" << std::endl;
  
  WDEBUG_OUT << L"executing statement" << std::endl;
  DataSet result( stmt.executeQuery() );
  WDEBUG_OUT << L"executed statement" << std::endl;
  
  IdList ids;
  for (result->next(); !result->isAfterLast(); result->next())
    ids.push_back( result->getLong(1) );
  
  WDEBUG_OUT << L"returning ids" << std::endl;
  return ids;
}

Data::ValueMap
Database:: get_properties(const Tk::String &domain, const Id &entity_id)
{
  
}

Data::TypeMap
Database:: get_domain_properties(const Tk::String &domain)
{
  const static Tk::String
    sql(" select   pt.name as property_type_name"
	" ,        pvt.name as property_value_type_name"
	" from     property_type pt"
	" ,        property_value_type pvt"
	" where    pt.domain                  = ?"
	" and      pvt.property_value_type_id = pt.property_value_type_id"
	" order by property_type_id"
	);
  PStatement *query( prepare_statement(sql) );
  
  query->setString(1, domain);
  
  DataSet result( query->executeQuery() );
  
  Data::TypeMap properties;
  for (result->next(); !result->isAfterLast(); result->next())
    properties[result->getString(1)] = Data::Value::get_type(result->getString(2));
	       //    properties.push_back( TypeMap::value_type( result->getString(1) );
  
  return properties;
}

Id
Database:: create_entity(const Tk::String &domain)
{
  Tk::StringList columns;
  columns.push_back(domain + Tk::String("_id"));
  
  PStatementPtr stmt( get_table_insert_statement(domain, columns) );
  
  stmt->execute();
  
  return get_last_insert_id();
}
