/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004, 2005, 2006, 2007 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <memory>

#include <odbc++/drivermanager.h>

#include "login.h"

using namespace Tk;
using namespace Db::Ui;


static const Tk::String login_conf_base("/apps/tkresolver/login/");

static Tk::String
get_conf_value(const Glib::RefPtr<Gnome::Conf::Client> &conf,
	       const Tk::String &value,
	       const Tk::String &default_value)
{
  Gnome::Conf::Value conf_value;
  Tk::String key(login_conf_base + value);

  try
    {
      conf_value = conf->get(key);
    }
  catch (...)
    {
      return default_value;
    }
  
  if (conf_value.get_string().length() < 1)
    return default_value;
  
  return static_cast<Tk::String>(conf_value.get_string());
}

Login:: Login()
  : Gtk::Dialog("Database login"),
    _conf(Gnome::Conf::Client::get_default_client())
{
  get_vbox()->set_spacing(4);
  set_has_separator();
  Gtk::Button * ok_button( add_button(Gtk::Stock::OK,     Gtk::RESPONSE_OK) );
  add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
  set_default_response(Gtk::RESPONSE_OK);

  _password.set_activates_default();
  _username.set_text(get_conf_value(_conf, "username", "psi"));
  _password.set_visibility(false);
  
  Gtk::Table * table = Gtk::manage(new Gtk::Table(4, 2));
  get_vbox()->pack_start(*table, true, true);
  table->set_spacings(4);
  
  table->attach(*Gtk::manage(new Gtk::Label("Data source", Gtk::ALIGN_LEFT)), 0, 1, 1, 2, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Username",    Gtk::ALIGN_LEFT)), 0, 1, 2, 3, Gtk::FILL, Gtk::FILL);
  table->attach(*Gtk::manage(new Gtk::Label("Password",    Gtk::ALIGN_LEFT)), 0, 1, 3, 4, Gtk::FILL, Gtk::FILL);
  
  table->attach(_data_source_name, 1, 2, 1, 2, Gtk::EXPAND|Gtk::FILL, Gtk::FILL);
  table->attach(_username, 1, 2, 2, 3, Gtk::EXPAND|Gtk::FILL, Gtk::FILL);
  table->attach(_password, 1, 2, 3, 4, Gtk::EXPAND|Gtk::FILL, Gtk::FILL);
  
  fill_data_source_names();
  
  _password.grab_focus();
  show_all_children();
}

Login:: ~Login()
{
  Tk::String conf(login_conf_base + Tk::String("data_source_name"));
  _conf->set(conf, get_data_source_name());
  
  conf = login_conf_base + Tk::String("username");
  _conf->set(conf, get_username());
}

void
Login:: fill_data_source_names()
{
  std::auto_ptr<odbc::DataSourceList> data_sources( odbc::DriverManager::getDataSources() );

  if (data_sources.get() != 0 && data_sources->size() > 0)
    {
      odbc::DataSourceList::iterator iter;
      iter = data_sources->begin();
      odbc::DataSourceList::iterator end;
      end = data_sources->end();
      for (;
	   iter != end;
	   ++iter)
	{
	  _data_source_name.append_text( String((*iter)->getName()) );
	}
      
      _data_source_name.set_active_text( String(data_sources->front()->getName()) );
    }
  
  _data_source_name.set_active_text( get_conf_value(_conf, "data_source_name", "TkResolver++") );
}

Tk::String
Login:: get_data_source_name() const
{
  return _data_source_name.get_active_text();
}

#define get_text(func, entry) \
Tk::String \
Login:: func() const \
{ \
  return entry.get_text(); \
}

get_text(get_username, _username);
get_text(get_password, _password);


Db::Database *
Tk::Db::Ui:: login()
{
    Db::Ui::Login * login( new Db::Ui::Login() );
    int login_response = login->run();
    if (login_response != Gtk::RESPONSE_OK)
      {
	delete login;
	return 0;
      }

    Tk::String data_source_name(login->get_data_source_name());
    Tk::String username(login->get_username());
    Tk::String password(login->get_password());
    delete login;
    
    if (data_source_name.size() < 1)
      {
	return 0;
      }
    
    return new Db::Database(data_source_name, username, password);
}
