/* -*- c++ -*-
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>
#include <errno.h>

#include "libtkcommon/debug.h"

#include "libtkquery/parse.h"
#include "libtkquery/parse_error.h"
#include "libtkquery/tkquery-parser.h"

using namespace Tk::Query;

extern int yy_flex_debug;
extern FILE *yyin;
extern void yyrestart(FILE *new_yyin);
Expression *expr_glbl;

static SyntaxError err_glbl(Location(), "no error");

// Mandatory error function
void yy::parser::error(const yy::parser::location_type& yylloc, const std::string& message)
{
  err_glbl = SyntaxError(yylloc, message);
}

static Expression *
do_parse(const Tk::String &domain)
{
  yyrestart(yyin);
  
  yy::parser *parser( new yy::parser() );
  //  parser->set_debug_level(1);
  int err( parser->parse() );
  
  delete parser;
  fclose(yyin);
  
  if (err)
    throw err_glbl;
  
  return expr_glbl;
}

Expression *
Tk::Query:: parse_file(const Tk::String &domain, const Tk::String &filename)
{
  yy_flex_debug = 1;
  
  yyin = fopen(std::string(filename).c_str(), "r");
  if (yyin == NULL)
    throw Tk::UnixError(std::string("Could not open query file '")
			+ static_cast<std::string>(filename)
			+ "'",
			errno);
  
  return do_parse(domain);
}

Expression *
Tk::Query:: parse_string(const Tk::String &domain, const Tk::String &query)
{
  std::string buf( query );
  
  yyin = fmemopen((void *)(buf.c_str()), strlen(buf.c_str()), "r");
  if (yyin == NULL)
    throw UnixError("Could not open memory as stream", errno);
  
  return do_parse(domain);
}


bool
Tk::Query:: query_is_valid(Tk::Db::Database *database, const Tk::String &domain, const Tk::String &query)
{
  try
    {
      ExpressionPtr expr( parse_string(domain, query) );
      QueryPtr query( expr->build_query(database, domain) );
    }
  catch (ParseError &e)
    {
      return false;
    }
  
  return true;
}

static Tk::Db::PStatement *
build_query(Tk::Db::Database *database, const Tk::String &domain, const Tk::String &filename)
{
  ExpressionPtr expr( parse_file(domain, filename) );
  QueryPtr query( expr->build_query(database, domain) );
  
  return query->get_statement(domain);
}
 
Tk::Db::IdList
Tk::Query:: get_experiments(Tk::Db::Database *database, const Tk::String &filename)
{
  Tk::Db::PStatement *stmt( build_query(database, "experiment", filename) );
  
  Tk::Db::DataSetPtr results( new Tk::Db::DataSet(stmt->executeQuery()) );
  
  bool valid( (*results)->next() );
  if (!valid)
    return Tk::Db::IdList();
  
  Tk::Db::IdList ids;
  for (;
       !(*results)->isAfterLast();
       (*results)->next())
    {
      ids.push_back((*results)->getLong(1));
    }
  
  return ids;
}
